//implementation for rational class

#include "rational.h"
using namespace std;

//default constructor - numerator 0, deominator 1
rational::rational()
	:numerator(0),denominator(1)
{}

//string constructor (for the calculator)
rational::rational(string rationalIn)
{
	//string vectorizing fn from http://stackoverflow.com/a/237280
	numerator = 0;
    denominator = 1;
	int mynumerator = 0;
	int mydenominator = 1;
	//parse the string
	std::vector<std::string> parsecons = split(rationalIn, '/');
	//set numerator
	if (parsecons.size() > 0) mynumerator = string2Int(parsecons[0]);
	//set denominator
	if (parsecons.size() > 1) mydenominator = string2Int(parsecons[1]);
	//build rational
	if ((mydenominator != 1) && (mydenominator != 0))
	{
	numerator = mynumerator;
	denominator = mydenominator;
	return;
	}
	else
	{
		if (mydenominator == 0) cerr << "Error: n/0 is undefined. Setting " << rationalIn << " to " << mynumerator << "/1." << endl;
		numerator = mynumerator;
	}

}

//constructor for n/n
rational::rational(int numeratorIn, int denominatorIn)
	:numerator(numeratorIn),denominator(denominatorIn)
{}

//constructor for n/1
rational::rational(int numeratorIn)
	:numerator(numeratorIn),denominator(1)
{}

//normalization
//Brian Overland:"C++ Without Fear: A Beginner's Guide That Makes You Feel Smart"
//http://books.google.ca/books?id=bW6MiHxPULUC&pg=PT504&lpg=PT504&dq=c%2B%2B+normalize+fraction&source=bl&ots=jdmWeRFAJR&sig=6P8AbmvID3fKUq1oaX9PIvoZ_fw&hl=en&sa=X&ei=ozZ8UNqRLK230gGe44HoDg&ved=0CFcQ6AEwCA
rational rational::normalize(rational toNormalize){
// Handle cases involving 0 (though denominator should never ever ever be 0, and it'd be more accurate to say such was undefined)
	if (toNormalize.numerator == 0 || toNormalize.denominator == 0) {
         toNormalize.numerator = 0;
		 toNormalize.denominator = 1;
	}
// Put neg. sign in numerator only.
	//this is actually what you want - if they were *both* negative, it's the same to have them both positive
    if (toNormalize.denominator < 0) {
         toNormalize.numerator *= -1;
		 toNormalize.denominator *= -1;
    }
// Factor out GCF from numerator and denominator.
    int gcfdiv = gcf(toNormalize.numerator, toNormalize.denominator);
        toNormalize.numerator = toNormalize.numerator / gcfdiv;
        toNormalize.denominator = toNormalize.denominator / gcfdiv;

    return toNormalize;
}

//greatest common factor (for normalization and LCM)
//ask to see my chicken scratch where I worked out why this works
int rational::gcf(int a, int b){
	if (a % b == 0)
		return abs(b);//but denominator shouldn't ever be 0
	else
		return gcf(b, a % b);
}

//least common multiple (for addition and subtraction - must have like denominators)
//info: http://en.wikipedia.org/wiki/Least_common_multiple#Reduction_by_the_greatest_common_divisor
int rational::lcm(int a, int b){
    return (a / gcf(a, b)) * b;
}

//compare functions: less than
bool rational::operator<(rational &r1){
	double first = double(this->numerator) / double(this->denominator);
	double second = double(r1.numerator) / double(r1.denominator);
	cout << "Testing if " + int2String(this->numerator) + "/" + int2String(this->denominator) + " < " + int2String(r1.numerator) + "/" + int2String(r1.denominator) << endl;
	if (first < second){ 
		cout << "Yes, it is!" << endl;
		return true;}
	else {
		cout << "I'm sorry, it isn't." << endl;
		return false;}
}

//compare functions: greater than
bool rational::operator>(rational &r1){
	double first = double(this->numerator) / double(this->denominator);
	double second = double(r1.numerator) / double(r1.denominator);
	cout << "Testing if " + int2String(this->numerator) + "/" + int2String(this->denominator) + " > " + int2String(r1.numerator) + "/" + int2String(r1.denominator) << endl;
	if (first > second){ 
		cout << "Yes, it is!" << endl;
		return true;}
	else {
		cout << "I'm sorry, it isn't." << endl;
		return false;}
}

//compare functions: equals
bool rational::operator==(rational &r1){
	double first = double(this->numerator) / double(this->denominator);
	double second = double(r1.numerator) / double(r1.denominator);
	cout << "Testing if " + int2String(this->numerator) + "/" + int2String(this->denominator) + " = " + int2String(r1.numerator) + "/" + int2String(r1.denominator) << endl;
	if (first == second){ 
		cout << "Yes, it does!" << endl;
		return true;}
	else {
		cout << "I'm sorry, it doesn't." << endl;
		return false;}
}

//input operator (out of class)
void operator>> (istream& input, rational &r1)
{
	//string vectorizing fn from http://stackoverflow.com/a/237280
	bool inputgood = false;
	string str;
	int denominator = 1;
	int numerator = 0;
	//can have 0 to 1 minus signs, must have 1-9 digits, followed by 0 to 1 of (0 to 1 /, 0 to 1 minus signs, then 1-9 digits)
	regex rgx("^-?[0-9]{1,9}(/-?[0-9]{1,9})?$");
	while (!inputgood && !cin.fail())
	{
	str = "";
	getline(input,str);
	if (input.eof()) return;//get out if someone inputs Ctrl+Z
	//check to make sure expression is valid
	if (!std::regex_match(str,rgx))
	{
		cerr << "\nInput failed! Please try again: ";
	}
	else{ inputgood = true;
	}
	}
	//parse the string
	std::vector<std::string> parsed = split(str, '/');
	//set numerator
	if (parsed.size() > 0) numerator = r1.string2Int(parsed[0]);
	//set denominator
	if (parsed.size() > 1){
	denominator = r1.string2Int(parsed[1]);
	}
	//build rational
	if ((denominator != 1) && (denominator != 0))
	{
	r1 = rational(numerator, denominator);
	return;
	}
	else
		if (denominator == 0) cerr << "Error: n/0 is undefined. Setting to n/1." << endl;
		r1 = rational(numerator);
}//end friend>>

//for making integers of strings
int rational::string2Int(string numberIn)
{
	int number;
	stringstream stream;
	stream << numberIn;
	stream >> number;
	return number;
}

//output operator (out of class)
ostream& operator<< (ostream &output, rational &r1)
{
	output << r1.toString(r1);
	return output;
}//end friend<<

//for output - makes strings of rationals
string rational::toString(rational r1){
	if (r1.denominator != 1)//anything over 1 is itself
	return int2String(r1.numerator) + "/" + int2String(r1.denominator);
	else return int2String(r1.numerator);
}

//for making strings of integers
string rational::int2String(int a)
{
	   stringstream s;
       s << a;
       string str = s.str();
	   return str;
}

//+ operator
//Brian Overland:"C++ Without Fear: A Beginner's Guide That Makes You Feel Smart"
//http://books.google.ca/books?id=bW6MiHxPULUC&pg=PT504&lpg=PT504&dq=c%2B%2B+normalize+fraction&source=bl&ots=jdmWeRFAJR&sig=6P8AbmvID3fKUq1oaX9PIvoZ_fw&hl=en&sa=X&ei=ozZ8UNqRLK230gGe44HoDg&ved=0CFcQ6AEwCA
rational rational::operator+(rational &r1){
	 int lcd = r1.lcm(this->denominator,r1.denominator);
	 int quot1 = lcd/this->denominator;
	 int quot2 = lcd/r1.denominator;
	 int a = this->numerator * quot1;
	 int b = r1.numerator * quot2;
	 if (!isAdditionSafe(a,b)){
		cerr << "Sorry, this operation would cause an overflow.\nReturning your first argument unchanged." << endl;
	    return r1;
	 }
	 rational sum(((this->numerator * quot1) + (r1.numerator * quot2)),lcd);
	 return sum.normalize(sum);
}

//- operator
rational rational::operator-(rational &r1){
	 int lcd = r1.lcm(this->denominator,r1.denominator);
	 int quot1 = lcd/this->denominator;
	 int quot2 = lcd/r1.denominator;
	 int a = this->numerator * quot1;
	 int b = r1.numerator * quot2;
	 if (!isAdditionSafe(a,(b * -1))){//since I'm effectively adding a negative when subtracting
		cerr << "Sorry, this operation would cause an overflow.\nReturning your first argument unchanged." << endl;
	    return r1;
	 }
	 rational difference(((this->numerator * quot1) - (r1.numerator * quot2)),lcd);
	 return difference.normalize(difference);
}

//* operator
rational rational::operator*(rational &r1){
	 int a = this->numerator;
	 int b = r1.numerator;
     int c = this->denominator;
	 int d = r1.denominator;
	 if (!isMultiplicationSafe(a,b) || !isMultiplicationSafe(c,d)){
	 	cerr << "Sorry, this operation would cause an overflow.\nReturning your first argument unchanged." << endl;
	    return r1;
	 }
	 rational product((this->numerator * r1.numerator), (this->denominator * r1.denominator));
	 return product.normalize(product);
}

//divide / operator
rational rational::operator/(rational &r1){
	if (r1.numerator == 0){
	cerr << "Error: Dividing by zero is undefined. Returning your dividend unchanged." << endl;
	return *this;
	}
	 int a = this->numerator;
	 int b = r1.numerator;
     int c = this->denominator;
	 int d = r1.denominator;
	 if (!isMultiplicationSafe(a,d) || !isMultiplicationSafe(b,c)){
	 	cerr << "Sorry, this operation would cause an overflow.\nReturning your first argument unchanged." << endl;
	    return r1;
	 }
	 rational quotient((this->numerator * r1.denominator), (this->denominator * r1.numerator));
	 return quotient.normalize(quotient);
}

//vectorizes a string by delimiter
//http://stackoverflow.com/a/237280
std::vector<std::string> &split(const std::string &s, char delim, std::vector<std::string> &elems) {
    std::stringstream ss(s);
    std::string item;
    while(std::getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}

//takes a string and a delimiter and calls the function to vectorize
//the string by delimiter, returns the split string
//http://stackoverflow.com/a/237280
std::vector<std::string> split(const std::string &s, char delim) {
    std::vector<std::string> elems;
    return split(s, delim, elems);
}

//"is...Safe" series to guard against integer overflow
//http://stackoverflow.com/a/1514309
//NOTE: I couldn't actually cause addition / subtraction overflow in this implementation because
//the lengths are restricted to 9 digits, and the max is high enough in the 10 digits to accomodate
//here in Microsoft Land it's +/- 2147483647
bool rational::isAdditionSafe(int a, int b){
	if ((b > 0) && (a > INT_MAX - b))
		return false;
    if ((b < 0) && (a < INT_MIN - b))
		return false;
	return true;
}

//...however the same is not true of multiplication (and division)
bool rational::isMultiplicationSafe(int a, int b){
	if ((b > 0) && (a > 0) && (a > INT_MAX / b))
		return false;
	else if (((b < 0) || (a < 0)) && (abs(a) > INT_MAX / abs(b)))
		return false;
	return true;
}

//virtual destructor
rational::~rational(){}

